unit uSaneScanner;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, StdCtrls, Sane;

{
 Code Flow
|------------------------------------------------------------------------------|
|  - sane_init()                                                               |
|    - pick designed device, possibly by using                                 |
|               sane_get_devices()                                             |
|    |-------------------------------------------------------------------------|
|    |  - sane_open()                                                          |
|    |    |--------------------------------------------------------------------|
|    |    |  - use:                                                            |
|    |    |         sane_get_option_descriptor()                               |
|    |    |         sane_control_option()                                      |
|    |    |    repeatedly to configure device as desired                       |
|    |    |--------------------------------------------------------------------|
|    |    |--------------------------------------------------------------------|
|    |    |  - sane_start()                                                    |
|    |    |  - use:                                                            |
|    |    |         sane_get_parameters()                                      |
|    |    |         sane_read()                                                |
|    |    |    repeatedly unit read returns EOF                                |
|    |    |  - go back to sane_start() if more frames desired                  |
|    |    |  - sane_cancel()                                                   |
|    |    |--------------------------------------------------------------------|
|    |  - sane_close()                                                         |
|    |-------------------------------------------------------------------------|
|  - sane_exit()                                                               |
|------------------------------------------------------------------------------|
}

type
  WordList  = array [0..128] of SANE_Word;
  PWordList = ^WordList;

type
  StrList   = array[0..255] of PChar;
  PStrList  = ^StrList;

type
  TSaneScanner = class
  private
    aScannerList : array of SANE_Device;
    bIsLibInit   : Boolean;
    bIsDevOpen   : Boolean;
    hScanner     : SANE_Handle;
    nDevice      : Integer;
    rInfo        : TMemo;
    sVersion     : String;
    LastStatus   : SANE_STATUS;
    function LastStatusOk: Boolean;
    procedure WriteInfo(aOption: Byte; aInfo: String);
    procedure SaneGetDeviceList;
    function SaneGetOptionCount: Integer;
    // Getter, Setter
    function GetVersion: String;
    function GetFoundDevices: Boolean;
    function GetDevice: Integer;
    procedure SetDevice(aValue: Integer);
    procedure SetMemo(aMemo: TMemo);
  public
    constructor Create;
    destructor Destroy;
    // Sane
    property SaneVersion: String read GetVersion;
    // Geräte
    property FoundDevices: Boolean read GetFoundDevices;
    property Device: Integer read GetDevice write SetDevice;
    // Ausgabe
    property Memo: TMemo write SetMemo;
    // Gerätelisten
    procedure DeviceList(aList: TStrings);
    procedure ScannerList(aList: TStrings);
  end;

implementation

uses
  dialogs, forms;

const
  asSaneValueType : Array[SANE_TYPE_BOOL..SANE_TYPE_GROUP] of String = ('Boolean', 'Integer', 'Fixed', 'String', 'Button', 'Group');
  asSaneUnitLong  : Array[SANE_UNIT_NONE..SANE_UNIT_MICROSECOND] of String = ('Keine', 'Pixel', 'Bit', 'Millimeter', 'Punkte je Zoll', 'Prozent', 'Microsecunden');
  asSaneUnit      : Array[SANE_UNIT_NONE..SANE_UNIT_MICROSECOND] of String = ('Keine', 'Pixel', 'Bit', 'mm', 'dpi', '%', 'µs');
  asOptionCap     : Array[False..True] of String = ('Nicht Aktiv', 'Aktiv');
  asOptionSet     : Array[False..True] of String = ('Nicht änderbar', 'Änderbar');

resourcestring
  rsCurrentValue           = 'Eingestellt';
  rsDevisesNotFound        = 'Es wurden keine Geräte gefunden.';
  rsDeviceName             = 'Geraetename';
  rsDevicesSearch          = 'Suche nach Scannern';
  rsEntries                = 'Eintraege';
  rsList                   = 'Liste';
  rsNoneCon                = 'Keine';
  rsOptionConstraintType   = 'Grenzen';
  rsOptionCount            = 'Anzahl der Optionen';
  rsOptionDescription      = 'Beschreibung';
  rsOptionName             = 'Name';
  rsOptionNummer           = 'Option Nummer';
  rsOptionOption           = 'Option';
  rsOptionPhysUnit         = 'Phys. Einheit';
  rsOptionTitle            = 'Titel';
  rsOptionSize             = 'Groesze';
  rsOptionValueType        = 'Wertetyp';
  rsRange                  = 'Bereich';
  rsRangeMax               = 'Maximum';
  rsRangeMin               = 'Minimum';
  rsRangeQuant             = 'Schrittweite';
  rsResult                 = 'Ergebnis';
  rsSelectDevice           = 'Ausgewaehltes Geraet';
  rsStringList             = 'Textliste';
  rsSystemName             = 'Systemname';
  rsValues                 = 'Werte';
  rsWordList               = 'Werteliste';

function TSaneScanner.LastStatusOk: Boolean;
{*******************************************************************************
* War der letzte Status in Ordnung?
}
begin
  Result := LastStatus = SANE_STATUS_GOOD;
end;

procedure TSaneScanner.WriteInfo(aOption: Byte; aInfo: String);
{*******************************************************************************
* Informationen in ein verbundenes Memo ausgeben.
}
begin
  if Assigned(rInfo)
  then begin
    case aOption
    of
      // Eintrag als neue Zeile
      0 : rInfo.Lines.Add(aInfo);
      // Eintrag als neue Zeile nach einer Leerzeile.
      1 : begin
            rInfo.Lines.Add('');
            rInfo.Lines.Add(aInfo);
          end;
      // Eintrag am Ende der letzten Zeile.
      2 : rInfo.Lines.Strings[rInfo.Lines.Count - 1] := rInfo.Lines.Strings[rInfo.Lines.Count - 1] + aInfo;
    end;
  end;
  Application.ProcessMessages;
end;

procedure TSaneScanner.SaneGetDeviceList;
{*******************************************************************************
*  Die Liste der verfügbaren Scanner holen.
}
var
  pdList : PDevice_List;
  i      : Integer;
  iVersion: SANE_Int;
begin
  // Library schliessen
  if bIsLibInit
  then sane_exit;
  // Neues Init
  bIsLibInit := (sane_init(@iVersion, nil) = SANE_STATUS_GOOD);

  if bIsLibInit
  then begin
    try
      WriteInfo(0, Format('LibSane Version: %s', [sVersion]));
      WriteInfo(1, rsDevicesSearch);
      i := 0;
      pdList := nil;
      SetLength(aScannerList, 0);
      Getmem(pdlist, sizeof(pdlist));
      LastStatus := sane_get_devices(pdlist, SANE_TRUE);
      if (LastStatus = SANE_STATUS_GOOD) then
      while (pdlist^[i] <> nil)
      do begin
        SetLength(aScannerList, i + 1);
        aScannerList[i].name     := pdlist^[i]^.name;     // Name im System
        aScannerList[i].vendor   := pdlist^[i]^.vendor;   // Hersteller
        aScannerList[i].model    := pdlist^[i]^.model;    // Modellbezeichnung
        aScannerList[i].dev_type := pdlist^[i]^.dev_type; // Bauart
        WriteInfo(1, Format('  %-11s : %s', [rsSystemName, aScannerList[i].name]));
        WriteInfo(0, Format('  %-11s : %s %s', [rsDeviceName, aScannerList[i].vendor, aScannerList[i].model]));
        Inc(i);
      end;
    finally
      if (pdList <> nil)
      then Freemem(pdList);
      if (i = 0)
      then WriteInfo(0, Format('  %s : %s', [rsResult, rsDevisesNotFound]));
    end;
  end;
end;


function TSaneScanner.SaneGetOptionCount: Integer;
{*******************************************************************************
* Anzahl der Optionen ermitteln
}
var
  SaneOption : PSANE_Option_Descriptor;
begin
  // Voreinstellung
  Result := -1;
  // Wenn ein Gerät ausgewählt ist ...
  if bIsDevOpen
  then begin
    Result := 0;
    // Beschreibungsliste holen
    SaneOption := sane_get_option_descriptor(hScanner, Result);
    // Die Anzahl en Index ermitteln.
    while (SaneOption <> nil)
    do begin
      inc(Result);
      SaneOption := sane_get_option_descriptor(hScanner, Result);
    end;
  end;
end;

// Getter, Setter

function TSaneScanner.GetVersion: String;
{*******************************************************************************
* Die unterstützte SaneVersion.
}
begin
  if bIsLibInit
  then Result := sVersion
  else Result := '';
end;

function TSaneScanner.GetFoundDevices: Boolean;
{*******************************************************************************
* Wurden Scanner gefunden?
}
begin
  // Geräte ermitteln
  SaneGetDeviceList;
  // Auswertung
  Result := Length(aScannerList) > 0;
end;

function TSaneScanner.GetDevice: Integer;
{*******************************************************************************
* Das ausgewählte Gerät.
}
begin
  Result := nDevice;
end;

procedure TSaneScanner.SetDevice(aValue: Integer);
{*******************************************************************************
* Die Nummer des aus der Liste gewählten Gerätes.
}
var
  SaneOption : PSANE_Option_Descriptor;
  pRange     : PSANE_Range;
  pWList     : PWordList;
  psList     : PStrList;
  pValue     : Pointer;
  iOption    : Integer;
  j, z       : integer;
begin
  if bIsLibInit and (nDevice <> aValue)
  then begin
    if (aValue >= 0) and (aValue < Length(aScannerList))
    then begin
      nDevice := aValue;
      // Ggfs Gerät schliessen
      if bIsDevOpen
      then begin
        sane_close(hScanner);
        hScanner := nil;
      end;
      // Trenner
      WriteInfo(1, '---------------------------');
      // Gewähltes gerät öffnen
      bIsDevOpen := (sane_open(PChar(aScannerList[nDevice].name), @hScanner) = SANE_STATUS_GOOD);
      // Ausgewählter Scanner
      WriteInfo(1, Format('%-20s : %s %s', [rsSelectDevice, aScannerList[aValue].vendor, aScannerList[aValue].model]));
      // Anzahl der Eigenschaften holen
      z := SaneGetOptionCount;
      WriteInfo(0, Format('%-20s : %d', [rsOptionCount, z]));
      // Die Eigenschaften holen
      if (z > 0) then
      for iOption := 0 to z - 1
      do begin
        // Kopf schreiben
        WriteInfo(1, Format('  %-15s   : %d', [rsOptionNummer, iOption]));
        // Die einzelne Eigenschaft holen
        SaneOption := sane_get_option_descriptor(hScanner, iOption);
        // Name
        WriteInfo(0, Format('    %-15s : "%s"', [rsOptionName, SaneOption^.name]));
        // Titel
        WriteInfo(0, Format('    %-15s : %s', [rsOptionTitle, SaneOption^.title]));
        // Beschreibung
        WriteInfo(0, Format('    %-15s : %s', [rsOptionDescription, SaneOption^.desc]));
        // Type des Wertes
        WriteInfo(0, Format('    %-15s : %s', [rsOptionValueType, asSaneValueType[SaneOption^.value_type]]));
        // Physiche Einheit
        WriteInfo(0, Format('    %-15s : %s (%s)', [rsOptionPhysUnit, asSaneUnitLong[SaneOption^.phys_unit], asSaneUnit[SaneOption^.phys_unit]]));
        // Größe des folgenden Werte
        WriteInfo(0, Format('    %-15s : %d', [rsOptionSize, SaneOption^.size]));
        // Eigenschaft aktiv
        WriteInfo(0, Format('    %-15s : %s', [rsOptionOption, asOptionCap[SANE_OPTION_IS_ACTIVE(SaneOption^.cap)]]));
        // Eigenschaft änderbar
        WriteInfo(0, Format('    %-15s : %s', [rsOptionOption, asOptionSet[SANE_OPTION_IS_SETTABLE(SaneOption^.cap)]]));
        // Einschränkungen der Werte
        case SaneOption^.constraint_type
        of
          SANE_CONSTRAINT_NONE:
          // Keine Beschränkung
          begin
            WriteInfo(0, Format('    %-15s : %s', [rsOptionConstraintType, rsNoneCon]));
          end;
          SANE_CONSTRAINT_RANGE:
          // Wertebereich
          begin
            WriteInfo(0, Format('    %-15s : %s', [rsOptionConstraintType, rsRange]));
            pRange := SaneOption^.prange;
            // Mindestwert
            if SaneOption^.value_type = SANE_TYPE_FIXED
            then WriteInfo(0, Format('      %-13s : %s', [rsRangeMin, FloatToStr(SANE_UNFIX(pRange^.min))]))
            else WriteInfo(0, Format('      %-13s : %d', [rsRangeMin, pRange^.min]));
            // Höchstwert
            if SaneOption^.value_type = SANE_TYPE_FIXED
            then WriteInfo(0, Format('      %-13s : %s', [rsRangeMax, FloatToStr(SANE_UNFIX(pRange^.max))]))
            else WriteInfo(0, Format('      %-13s : %d', [rsRangeMax, pRange^.max]));
            // Schrittweite
            if SaneOption^.value_type = SANE_TYPE_FIXED
            then WriteInfo(0, Format('      %-13s : %s', [rsRangeQuant, FloatToStr(SANE_UNFIX(pRange^.quant))]))
            else WriteInfo(0, Format('      %-13s : %d', [rsRangeQuant, pRange^.quant]));
          end;
          // Werteliste
          SANE_CONSTRAINT_WORD_LIST:
          begin
            WriteInfo(0, Format('    %-15s : %s', [rsOptionConstraintType, rsWordList]));
            pwList := pWordlist(SaneOption^.pwordlist);
            // Liste eintragen
            WriteInfo(0, Format('      %-13s : %d %s', [rsList, pwList^[0], rsEntries]));
            WriteInfo(0, Format('      %-13s : ', [rsValues]));
            for j := 1 to pwList^[0] do
            if SaneOption^.value_type = SANE_TYPE_FIXED
            then WriteInfo(2, Format(' %s,', [FloatToStr(SANE_UNFIX(pwList^[j]))]))
            else WriteInfo(2, Format(' %d,', [pwList^[j]]))
          end;
          // TextListe
          SANE_CONSTRAINT_STRING_LIST:
          begin
            WriteInfo(0, Format('    %-15s : %s', [rsOptionConstraintType, rsStringList]));
            // Liste eintragen
            WriteInfo(0, Format('      %-13s :', [rsEntries]));
            psList := PStrList(SaneOption^.pstringlist);
            j      := 0;
            while (psList^[j] <> nil)
            do begin
              WriteInfo(2, Format(' %s,', [psList^[j]]));
              Inc(j);
            end;
          end;
        end;
        // Den aktuell eingestellten Wert ermitteln
        Getmem(pValue, SaneOption^.size);
        if (pValue <> nil) then
        try
          if (sane_control_option(hScanner, iOption, SANE_ACTION_GET_VALUE, pValue, nil) = SANE_STATUS_GOOD) then
          case SaneOption^.value_type of
            SANE_TYPE_BOOL   : WriteInfo(0, Format('      %-13s : %s', [rsCurrentValue, BoolToStr(Integer(pValue^) <> 0)]));
            SANE_TYPE_FIXED  : WriteInfo(0, Format('      %-13s : %s', [rsCurrentValue, FloatToStr(SANE_UNFIX(Integer(pValue^)))]));
            SANE_TYPE_INT    : WriteInfo(0, Format('      %-13s : $d', [rsCurrentValue, Integer(pValue^)]));
            SANE_TYPE_STRING : WriteInfo(0, Format('      %-13s : %s', [rsCurrentValue, PChar(pValue)]));
          end;
        finally
          Freemem(pValue);
        end;
      end;
    end;
  end;
end;

procedure TSaneScanner.SetMemo(aMemo: TMemo);
{*******************************************************************************
* Ein Memo zur Ausgabe von Informatioen.
}
begin
  rInfo := aMemo as TMemo;
  rInfo.Clear;
end;

constructor TSaneScanner.Create;
{*******************************************************************************
* Klasse erstellen und die Version der unterstützten Sane-Version ermitteln.
}
var
  iVersion: SANE_Int;
begin
  inherited Create;
  // Initialisieren.
  bIsLibInit := (sane_init(@iVersion, nil) = SANE_STATUS_GOOD);
  // Voreinstellung
  hScanner   := nil;
  nDevice    := -1;
  sVersion   := '';
  bIsDevOpen := False;
  // SaneVersion
  if bIsLibInit
  then sVersion := IntToStr(SANE_VERSION_MAJOR(iVersion) ) + '.'
                 + IntToStr(SANE_VERSION_MINOR(iVersion) ) + '.'
                 + IntToStr(SANE_VERSION_BUILD(iVersion) );
end;

destructor TSaneScanner.Destroy;
{*******************************************************************************
* Klasse beenden.
}
begin
  // Gerät schließen?
  if bIsDevOpen
  then sane_close(hScanner);
  // Library verlassen?
  if bIsLibInit
  then sane_exit;
  inherited Destroy;
end;

procedure TSaneScanner.DeviceList(aList: TStrings);
{*******************************************************************************
* Eine Liste der gefundenen Gerätenamen erstellen.
}
var
  i : Integer;
begin
  aList.Clear;
  if (Length(aScannerList) > 0)
  then begin
    for i := 0 to Length(aScannerList) - 1
    do aList.Add(aScannerList[i].name);
  end;
end;

procedure TSaneScanner.ScannerList(aList: TStrings);
{*******************************************************************************
* Eine Liste der gefundenen Scannermodelle erstellen.
}
var
  i : Integer;
begin
  aList.Clear;
  if (Length(aScannerList) > 0)
  then begin
    for i := 0 to Length(aScannerList) - 1
    do aList.Add(aScannerList[i].vendor + ' ' + aScannerList[i].model);
  end;
end;

end.

